from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Any

import httpx
import openai
from openai.types.chat import ChatCompletionChunk, ChatCompletionToolChoiceOptionParam
from openai.types.chat.chat_completion_chunk import Choice

from livekit.agents import APIConnectionError, APIStatusError, APITimeoutError, llm
from livekit.agents.llm import ToolChoice
from livekit.agents.llm.chat_context import ChatContext
from livekit.agents.llm.tool_context import FunctionTool
from livekit.agents.types import (
    DEFAULT_API_CONNECT_OPTIONS,
    NOT_GIVEN,
    APIConnectOptions,
    NotGivenOr,
)
from livekit.agents.utils import is_given

from .log import logger
from .utils import to_chat_ctx, to_fnc_ctx


@dataclass
class _LLMOptions:
    model: str
    user: NotGivenOr[str]
    temperature: NotGivenOr[float]
    parallel_tool_calls: NotGivenOr[bool]
    tool_choice: NotGivenOr[ToolChoice]
    store: NotGivenOr[bool]
    metadata: NotGivenOr[dict[str, str]]


class LLM(llm.LLM):
    def __init__(
        self,
        *,
        model: str = "MiniMax-Text-01",
        api_key: NotGivenOr[str] = NOT_GIVEN,
        client: openai.AsyncClient | None = None,
        user: NotGivenOr[str] = NOT_GIVEN,
        temperature: NotGivenOr[float] = NOT_GIVEN,
        parallel_tool_calls: NotGivenOr[bool] = NOT_GIVEN,
        tool_choice: NotGivenOr[ToolChoice] = NOT_GIVEN,
        store: NotGivenOr[bool] = NOT_GIVEN,
        metadata: NotGivenOr[dict[str, str]] = NOT_GIVEN,
        timeout: httpx.Timeout | None = None,
    ) -> None:
        """
        Create a new instance of Volcengine LLM.

        Args:
            model: The model to use for the LLM, end
            api_key: The API key to use for the LLM.
            base_url: The base URL to use for the LLM.
        """
        super().__init__()
        self._opts = _LLMOptions(
            model=model,
            user=user,
            temperature=temperature,
            parallel_tool_calls=parallel_tool_calls,
            tool_choice=tool_choice,
            store=store,
            metadata=metadata,
        )
        api_key = api_key if is_given(api_key) else os.getenv("MINIMAX_API_KEY")
        if api_key is None:
            raise ValueError("MINIMAX_API_KEY is required")
        self._client = client or openai.AsyncClient(
            api_key=api_key if is_given(api_key) else None,
            base_url="https://api.minimaxi.com/v1",
            max_retries=0,
            http_client=httpx.AsyncClient(
                timeout=timeout
                if timeout
                else httpx.Timeout(connect=15.0, read=5.0, write=5.0, pool=5.0),
                follow_redirects=True,
                limits=httpx.Limits(
                    max_connections=50,
                    max_keepalive_connections=50,
                    keepalive_expiry=120,
                ),
            ),
        )

    def chat(
        self,
        *,
        chat_ctx: ChatContext,
        tools: list[FunctionTool] | None = None,
        conn_options: APIConnectOptions = DEFAULT_API_CONNECT_OPTIONS,
        parallel_tool_calls: NotGivenOr[bool] = NOT_GIVEN,
        tool_choice: NotGivenOr[ToolChoice] = NOT_GIVEN,
        extra_kwargs: NotGivenOr[dict[str, Any]] = NOT_GIVEN,
    ) -> LLMStream:
        extra = {}

        if is_given(extra_kwargs):
            extra.update(extra_kwargs)

        if is_given(self._opts.metadata):
            extra["metadata"] = self._opts.metadata

        if is_given(self._opts.user):
            extra["user"] = self._opts.user

        parallel_tool_calls = (
            parallel_tool_calls
            if is_given(parallel_tool_calls)
            else self._opts.parallel_tool_calls
        )
        if is_given(parallel_tool_calls):
            extra["parallel_tool_calls"] = parallel_tool_calls

        tool_choice = tool_choice if is_given(tool_choice) else self._opts.tool_choice  # type: ignore
        if is_given(tool_choice):
            oai_tool_choice: ChatCompletionToolChoiceOptionParam
            if isinstance(tool_choice, dict):
                oai_tool_choice = {
                    "type": "function",
                    "function": {"name": tool_choice["function"]["name"]},
                }
                extra["tool_choice"] = oai_tool_choice
            elif tool_choice in ("auto", "required", "none"):
                oai_tool_choice = tool_choice
                extra["tool_choice"] = oai_tool_choice
        logger.info("llm start", extra={"model": self._opts.model})
        return LLMStream(
            self,
            model=self._opts.model,
            client=self._client,
            chat_ctx=chat_ctx,
            tools=tools or [],
            conn_options=conn_options,
            extra_kwargs=extra,
        )


class LLMStream(llm.LLMStream):
    def __init__(
        self,
        llm: LLM,
        *,
        model: str,
        client: openai.AsyncClient,
        chat_ctx: llm.ChatContext,
        tools: list[FunctionTool],
        conn_options: APIConnectOptions,
        extra_kwargs: dict[str, Any],
    ) -> None:
        super().__init__(llm, chat_ctx=chat_ctx, tools=tools, conn_options=conn_options)
        self._model = model
        self._client = client
        self._llm = llm
        self._extra_kwargs = extra_kwargs

    async def _run(self) -> None:
        # current function call that we're waiting for full completion (args are streamed)
        # (defined inside the _run method to make sure the state is reset for each run/attempt)
        self._oai_stream: openai.AsyncStream[ChatCompletionChunk] | None = None
        self._tool_call_id: str | None = None
        self._fnc_name: str | None = None
        self._fnc_raw_arguments: str | None = None
        self._tool_index: int | None = None
        retryable = True
        first_response = True
        try:
            stream: openai.AsyncStream[
                ChatCompletionChunk
            ] = await self._client.chat.completions.create(
                messages=to_chat_ctx(self._chat_ctx, id(self._llm)),
                tools=to_fnc_ctx(self._tools) if self._tools else openai.NOT_GIVEN,
                model=self._model,
                stream_options={"include_usage": True},
                stream=True,
                **self._extra_kwargs,
            )

            async with stream:
                async for chunk in stream:
                    for choice in chunk.choices:
                        chat_chunk = self._parse_choice(chunk.id, choice)
                        if chat_chunk is not None:
                            retryable = False
                            self._event_ch.send_nowait(chat_chunk)
                        if first_response:
                            logger.info("llm first response")
                            first_response = False

                    if chunk.usage is not None:
                        retryable = False
                        chunk = llm.ChatChunk(
                            id=chunk.id,
                            usage=llm.CompletionUsage(
                                completion_tokens=chunk.usage.completion_tokens,
                                prompt_tokens=chunk.usage.prompt_tokens,
                                total_tokens=chunk.usage.total_tokens,
                            ),
                        )
                        self._event_ch.send_nowait(chunk)
            logger.info("llm end")

        except openai.APITimeoutError:
            raise APITimeoutError(retryable=retryable)  # noqa: B904
        except openai.APIStatusError as e:
            raise APIStatusError(  # noqa: B904
                e.message,
                status_code=e.status_code,
                request_id=e.request_id,
                body=e.body,
                retryable=retryable,
            )
        except Exception as e:
            raise APIConnectionError(retryable=retryable) from e

    def _parse_choice(self, id: str, choice: Choice) -> llm.ChatChunk | None:
        delta = choice.delta

        if delta is None:
            return None

        if delta.tool_calls:
            for tool in delta.tool_calls:
                if not tool.function:
                    continue

                call_chunk = None
                if self._tool_call_id and tool.id and tool.index != self._tool_index:
                    call_chunk = llm.ChatChunk(
                        id=id,
                        delta=llm.ChoiceDelta(
                            role="assistant",
                            content=delta.content,
                            tool_calls=[
                                llm.FunctionToolCall(
                                    arguments=self._fnc_raw_arguments or "",
                                    name=self._fnc_name or "",
                                    call_id=self._tool_call_id or "",
                                )
                            ],
                        ),
                    )
                    self._tool_call_id = self._fnc_name = self._fnc_raw_arguments = None

                if tool.function.name:
                    self._tool_index = tool.index
                    self._tool_call_id = tool.id
                    self._fnc_name = tool.function.name
                    self._fnc_raw_arguments = tool.function.arguments or ""
                elif tool.function.arguments:
                    self._fnc_raw_arguments += tool.function.arguments  # type: ignore

                if call_chunk is not None:
                    return call_chunk

        if choice.finish_reason in ("tool_calls", "stop") and self._tool_call_id:
            call_chunk = llm.ChatChunk(
                id=id,
                delta=llm.ChoiceDelta(
                    role="assistant",
                    content=delta.content,
                    tool_calls=[
                        llm.FunctionToolCall(
                            arguments=self._fnc_raw_arguments or "",
                            name=self._fnc_name or "",
                            call_id=self._tool_call_id or "",
                        )
                    ],
                ),
            )
            self._tool_call_id = self._fnc_name = self._fnc_raw_arguments = None
            return call_chunk

        return llm.ChatChunk(
            id=id,
            delta=llm.ChoiceDelta(content=delta.content, role="assistant"),
        )



from __future__ import annotations

import json
import os
import time
from typing import Dict, Literal

import aiohttp
from pydantic import BaseModel, Field
from osc_data.text_stream import TextStreamSentencizer

from livekit.agents import (
    APIConnectOptions,
    tts,
    utils,
)
from livekit.agents.types import DEFAULT_API_CONNECT_OPTIONS

from .log import logger


class TTSOptions(BaseModel):
    api_key: str
    group_id: str
    base_url: str = "https://api.minimax.chat/v1/t2a_v2"
    model: Literal[
        "speech-02-hd",
        "speech-02-turbo",
        "speech-01-hd",
        "speech-01-turbo",
        "speech-01-240228",
        "speech-01-turbo-240228",
    ] = "speech-02-hd"
    language_boost: Literal[
        "Chinese",
        "Chinese,Yue",
        "English",
        "Arabic",
        "Russian",
        "Spanish",
        "French",
        "Portuguese",
        "German",
        "Turkish",
        "Dutch",
        "Ukrainian",
        "Vietnamese",
        "Indonesian",
        "Japanese",
        "Italian",
        "Korean",
        "Thai",
        "Polish",
        "Romanian",
        "Greek",
        "Czech",
        "Finnish",
        "Hindi",
        "auto",
    ] = "auto"

    # audio
    sample_rate: Literal[8000, 16000, 22050, 24000, 32000, 44100] = 16000
    bitrate: Literal[32000, 64000, 128000, 256000] = 128000  # 该参数仅对mp3格式有效
    audio_format: Literal["mp3", "pcm", "flac", "wav"] = "pcm"
    num_channels: Literal[1, 2] = 1

    # speech
    speed: float = Field(1.0, ge=0.5, le=2.0)
    volume: float = Field(1.0, gt=0.0, le=10.0)
    pitch: int = Field(0, ge=-12, le=12)
    voice_id: Literal[
        "male-qn-qingse",
        "male-qn-jingying",
        "male-qn-badao",
        "male-qn-daxuesheng",
        "female-shaonv",
        "female-yujie",
        "female-chengshu",
        "female-tianmei",
        "presenter_male",
        "presenter_female",
        "audiobook_male_1",
        "audiobook_male_2",
        "audiobook_female_1",
        "audiobook_female_2",
        "male-qn-qingse-jingpin",
        "male-qn-jingying-jingpin",
        "male-qn-badao-jingpin",
        "male-qn-daxuesheng-jingpin",
        "female-shaonv-jingpin",
        "female-yujie-jingpin",
        "female-chengshu-jingpin",
        "female-tianmei-jingpin",
        "clever_boy",
        "cute_boy",
        "lovely_girl",
        "cartoon_pig",
        "bingjiao_didi",
        "junlang_nanyou",
        "chunzhen_xuedi",
        "lengdan_xiongzhang",
        "badao_shaoye",
        "tianxin_xiaoling",
        "qiaopi_mengmei",
        "wumei_yujie",
        "diadia_xuemei",
        "danya_xuejie",
        "Santa_Claus",
        "Grinch",
        "Rudolph",
        "Arnold",
        "Charming_Santa",
        "Charming_Lady",
        "Sweet_Girl",
        "Cute_Elf",
        "Attractive_Girl",
        "Serene_Woman",
    ] = "male-qn-jingying"

    def get_http_url(self):
        return f"{self.base_url}?GroupId={self.group_id}"

    def get_http_header(self):
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        return headers

    def get_query_params(self, text: str) -> Dict:
        request_json = {
            "model": "speech-02-turbo",
            "text": text,
            "stream": True,
            "language_boost": self.language_boost,
            "output_format": "hex",
            "voice_setting": {
                "voice_id": self.voice_id,
                "speed": self.speed,
                "vol": self.volume,
                "pitch": self.pitch,
            },
            "audio_setting": {
                "sample_rate": self.sample_rate,
                "bitrate": self.bitrate,
                "format": self.audio_format,
            },
        }
        return request_json


class TTS(tts.TTS):
    def __init__(
        self,
        api_key: str | None = None,
        group_id: str | None = None,
        model: Literal[
            "speech-02-hd",
            "speech-02-turbo",
            "speech-01-hd",
            "speech-01-turbo",
            "speech-01-240228",
            "speech-01-turbo-240228",
        ] = "speech-02-turbo",
        language_boost: Literal[
            "Chinese",
            "Chinese,Yue",
            "English",
            "Arabic",
            "Russian",
            "Spanish",
            "French",
            "Portuguese",
            "German",
            "Turkish",
            "Dutch",
            "Ukrainian",
            "Vietnamese",
            "Indonesian",
            "Japanese",
            "Italian",
            "Korean",
            "Thai",
            "Polish",
            "Romanian",
            "Greek",
            "Czech",
            "Finnish",
            "Hindi",
            "auto",
        ] = "auto",
        sample_rate: Literal[8000, 16000, 22050, 24000, 32000, 44100] = 8000,
        bitrate: Literal[
            32000, 64000, 128000, 256000
        ] = 128000,  # 该参数仅对mp3格式有效
        audio_format: Literal["mp3", "pcm", "flac", "wav"] = "pcm",
        num_channels: Literal[1, 2] = 1,
        speed: float = 1.0,
        volume: float = 1.0,
        pitch: float = 0.0,
        voice_id: Literal[
            "male-qn-qingse",
            "male-qn-jingying",
            "male-qn-badao",
            "male-qn-daxuesheng",
            "female-shaonv",
            "female-yujie",
            "female-chengshu",
            "female-tianmei",
            "presenter_male",
            "presenter_female",
            "audiobook_male_1",
            "audiobook_male_2",
            "audiobook_female_1",
            "audiobook_female_2",
            "male-qn-qingse-jingpin",
            "male-qn-jingying-jingpin",
            "male-qn-badao-jingpin",
            "male-qn-daxuesheng-jingpin",
            "female-shaonv-jingpin",
            "female-yujie-jingpin",
            "female-chengshu-jingpin",
            "female-tianmei-jingpin",
            "clever_boy",
            "cute_boy",
            "lovely_girl",
            "cartoon_pig",
            "bingjiao_didi",
            "junlang_nanyou",
            "chunzhen_xuedi",
            "lengdan_xiongzhang",
            "badao_shaoye",
            "tianxin_xiaoling",
            "qiaopi_mengmei",
            "wumei_yujie",
            "diadia_xuemei",
            "danya_xuejie",
            "Santa_Claus",
            "Grinch",
            "Rudolph",
            "Arnold",
            "Charming_Santa",
            "Charming_Lady",
            "Sweet_Girl",
            "Cute_Elf",
            "Attractive_Girl",
            "Serene_Woman",
        ] = "male-qn-jingying",
        http_session: aiohttp.ClientSession | None = None,
        max_session_duration: float = 600,
    ):
        """mininax TTS

        Args:
            api_key (str | None, optional):  API key. Defaults to None.
            group_id (str | None, optional):  Group ID. Defaults to None.
            base_url (str | None, optional):  Base URL. Defaults to None.
            model (Literal[ "speech-02-hd", "speech-02-turbo", "speech-01-hd", "speech-01-turbo", "speech-01-240228", "speech-01-turbo-240228" ], optional):  Model. Defaults to "speech-02-hd".

        """
        super().__init__(
            capabilities=tts.TTSCapabilities(streaming=True),
            sample_rate=sample_rate,
            num_channels=1,
        )
        if api_key is None:
            api_key = os.environ.get("MINIMAX_API_KEY")
        if group_id is None:
            group_id = os.environ.get("MINIMAX_GROUP_ID")
        if api_key is None or group_id is None:
            raise ValueError("MINIMAX_API_KEY and MINIMAX_GROUP_ID must be provided")
        self._opts = TTSOptions(
            api_key=api_key,
            group_id=group_id,
            model=model,
            language_boost=language_boost,
            sample_rate=sample_rate,
            bitrate=bitrate,
            audio_format=audio_format,
            num_channels=num_channels,
            speed=speed,
            volume=volume,
            pitch=pitch,
            voice_id=voice_id,
        )
        self._session = http_session

    def _ensure_session(self) -> aiohttp.ClientSession:
        if self._session is None:
            self._session = utils.http_context.http_session()

        return self._session

    def synthesize(
        self, text, *, conn_options: APIConnectOptions = DEFAULT_API_CONNECT_OPTIONS
    ):
        raise NotImplementedError("Minimax TTS does not support synthesize method")

    def stream(self, *, conn_options: APIConnectOptions = DEFAULT_API_CONNECT_OPTIONS):
        return SynthesizeStream(
            tts=self,
            conn_options=conn_options,
            opts=self._opts,
            session=self._ensure_session(),
        )


class SynthesizeStream(tts.SynthesizeStream):
    def __init__(
        self,
        *,
        tts: TTS,
        opts: TTSOptions,
        session: aiohttp.ClientSession,
        conn_options: APIConnectOptions = DEFAULT_API_CONNECT_OPTIONS,
    ):
        super().__init__(tts=tts, conn_options=conn_options)
        self._opts, self._session = opts, session

    async def _run(self, emitter: tts.AudioEmitter) -> None:
        request_id = utils.shortuuid()
        emitter.initialize(
            request_id=request_id,
            sample_rate=self._opts.sample_rate,
            mime_type="audio/pcm",
            stream=True,
            num_channels=1,
        )
        splitter = TextStreamSentencizer()
        first_sentence_spend = None
        start_time = time.perf_counter()
        async for token in self._input_ch:
            if isinstance(token, self._FlushSentinel):
                sentences = splitter.flush()
            else:
                sentences = splitter.push(text=token)

            for sentence in sentences:
                if len(sentence.strip()) > 0:
                    if first_sentence_spend is None:
                        first_sentence_spend = time.perf_counter() - start_time
                        logger.info(
                            "llm first sentence",
                            extra={"spent": str(first_sentence_spend)},
                        )
                    emitter.start_segment(segment_id=utils.shortuuid())
                    first_response_spend = None
                    logger.info("tts start", extra={"sentence": sentence})
                    data = self._opts.get_query_params(text=sentence)
                    if first_response_spend is None:
                        start_time = time.perf_counter()
                    async with self._session.post(
                        self._opts.get_http_url(),
                        json=data,
                        timeout=aiohttp.ClientTimeout(
                            total=300,
                            sock_connect=self._conn_options.timeout,
                        ),
                        headers=self._opts.get_http_header(),
                    ) as resp:
                        resp.raise_for_status()
                        resp.content._high_water = (
                            resp.content._high_water**2
                        )  # 可能会出现内存占用过高的情况，暂时没找到更好方法
                        async for data in resp.content:
                            if data[:5] == b"data:":
                                data = json.loads(data[5:])
                                if "data" in data and "extra_info" not in data:
                                    audio = data["data"]["audio"]
                                    if first_response_spend is None:
                                        first_response_spend = (
                                            time.perf_counter() - start_time
                                        )
                                        logger.info(
                                            "tts first response",
                                            extra={"spent": str(first_response_spend)},
                                        )
                                    # audio hex编码
                                    audio = bytes.fromhex(audio)
                                    emitter.push(audio)
                    emitter.end_segment()
                    self._pushed_text = self._pushed_text.replace(sentence, "")
                    logger.info("tts end")



from __future__ import annotations

import base64
import os
from collections import OrderedDict
from collections.abc import Awaitable
from typing import Any, Callable, Union

from openai.types.chat import (
    ChatCompletionContentPartParam,
    ChatCompletionMessageParam,
    ChatCompletionToolParam,
)

from livekit.agents import llm

AsyncAzureADTokenProvider = Callable[[], Union[str, Awaitable[str]]]


def get_base_url(base_url: str | None) -> str:
    if not base_url:
        base_url = os.getenv(
            "VOLCENGINE_LLM_BASE_URL", "https://ark.cn-beijing.volces.com/api/v3"
        )
    return base_url


def to_fnc_ctx(fnc_ctx: list[llm.FunctionTool]) -> list[ChatCompletionToolParam]:
    return [llm.utils.build_strict_openai_schema(fnc) for fnc in fnc_ctx]


def to_chat_ctx(
    chat_ctx: llm.ChatContext, cache_key: Any
) -> list[ChatCompletionMessageParam]:
    # group the message and function_calls
    item_groups: dict[str, list[llm.ChatItem]] = OrderedDict()
    for item in chat_ctx.items:
        if (
            item.type == "message" and item.role == "assistant"
        ) or item.type == "function_call":
            group_id = item.id.split("/")[0]
            if group_id not in item_groups:
                item_groups[group_id] = []
            item_groups[group_id].append(item)
        else:
            item_groups[item.id] = [item]

    return [_group_to_chat_item(items, cache_key) for items in item_groups.values()]


def _group_to_chat_item(
    items: list[llm.ChatItem], cache_key: Any
) -> ChatCompletionMessageParam:
    if len(items) == 1:
        return _to_chat_item(items[0], cache_key)
    else:
        msg = {"role": "assistant", "tool_calls": []}
        for item in items:
            if item.type == "message":
                assert item.role == "assistant", (
                    "only assistant messages can be grouped"
                )
                assert "content" not in msg, (
                    "only one assistant message is allowed in a group"
                )

                msg.update(_to_chat_item(item, cache_key))
            elif item.type == "function_call":
                msg["tool_calls"].append(
                    {
                        "id": item.call_id,
                        "type": "function",
                        "function": {"name": item.name, "arguments": item.arguments},
                    }
                )
        return msg


def _to_chat_item(msg: llm.ChatItem, cache_key: Any) -> ChatCompletionMessageParam:
    if msg.type == "message":
        list_content: list[ChatCompletionContentPartParam] = []
        text_content = ""
        for content in msg.content:
            if isinstance(content, str):
                if text_content:
                    text_content += "\n"
                text_content += content
            elif isinstance(content, llm.ImageContent):
                list_content.append(_to_image_content(content, cache_key))

        if not list_content:
            # certain providers require text-only content in a string vs a list.
            # for max-compatibility, we will combine all text content into a single string.
            return {
                "role": msg.role,  # type: ignore
                "content": text_content,
            }

        if text_content:
            list_content.append({"type": "text", "text": text_content})

        return {
            "role": msg.role,  # type: ignore
            "content": list_content,
        }

    elif msg.type == "function_call":
        return {
            "role": "assistant",
            "tool_calls": [
                {
                    "id": msg.call_id,
                    "type": "function",
                    "function": {
                        "name": msg.name,
                        "arguments": msg.arguments,
                    },
                }
            ],
        }

    elif msg.type == "function_call_output":
        return {
            "role": "tool",
            "tool_call_id": msg.call_id,
            "content": msg.output,
        }


def _to_image_content(
    image: llm.ImageContent, cache_key: Any
) -> ChatCompletionContentPartParam:
    img = llm.utils.serialize_image(image)
    if img.external_url:
        return {
            "type": "image_url",
            "image_url": {
                "url": img.external_url,
                "detail": img.inference_detail,
            },
        }
    if cache_key not in image._cache:
        image._cache[cache_key] = img.data_bytes
    b64_data = base64.b64encode(image._cache[cache_key]).decode("utf-8")
    return {
        "type": "image_url",
        "image_url": {
            "url": f"data:{img.mime_type};base64,{b64_data}",
            "detail": img.inference_detail,
        },
    }
